#!/usr/bin/env python
"""
    Program:    vert.py
    License:    GPL 3
    Version:    0.8dev
    Last edit:  2012-11-14
    Desc:       Very simple 'transcoder' program for Linux.  Uses ImageMagick
                and ffmpeg to convert images, video and audio.  Has very few
                arguments; usage:  vert.py <input> <output>
                I/O is determined by file extension.
"""
import os, sys
import subprocess, time, shutil

##  These tuples are accepted extension for the three different file types
IMG_EXT = ('png','gif','bmp','jpeg','jpg','tiff')
AUD_EXT = ('mp3','wav','ogg','oga','flac','speex','spx')
VID_EXT = ('flv','mp4','m4v','mpg','avi','ogv','webm','mov','vob','mkv')


def set_file_options():
    """
    Sets intitial file options
    """
    fileOpts = {"clipStart":0,
            "clipDuration":0, "inFileExtension":"", "inFileFormat":"",
            "outFileExtension":"","outFileFormat":"", "inFile":"","outFile":"",
            "videoBitrate":"","audioCodec":"","noVideo":"","fps":0,
            "inFileType":"","outFileType":"","videoCodec":"",
            "program":"", "size":"", "audioRate":0, "audioChannels":0,
            "optimizeLayers":"", "delay":0, "imgCount":0}

    return fileOpts


def construct_system_call(fileOpts):
    """
    This function will 'build' a system call depending on options set.
    """
    if fileOpts["program"] == "ffmpeg":
        sysCall = "ffmpeg "
        if fileOpts["safe"] == False:
            sysCall += "-y "

        sysCall += "-i %s " % fileOpts["inFile"] + " "

        if fileOpts["noVideo"] == True:
            sysCall += "-vn "
        if fileOpts["audioCodec"] != "":
            sysCall += "-acodec %s " % fileOpts["audioCodec"]
        if fileOpts["audioRate"] != 0:
            sysCall += "-ab %d " % fileOpts["audioRate"]
        if fileOpts["audioChannels"] != 0:
            sysCall += "-ac %d " % fileOpts["audioChannels"]
        if fileOpts["outFileFormat"] != "":
            sysCall += "-f %s " % fileOpts["outFileFormat"]
        if fileOpts["videoCodec"] != "":
            sysCall += "-vcodec %s " % fileOpts["videoCodec"]
        if fileOpts["videoBitrate"] != "":
            sysCall += "-b:v %s " % fileOpts["videoBitrate"]
        if fileOpts["size"] != "":
            sysCall += "-s %s " % fileOpts["size"]
        if fileOpts["clipStart"] != 0:
            sysCall += "-ss %d " % fileOpts["clipStart"]
        if fileOpts["clipDuration"] != 0:
            sysCall += "-t %d " % fileOpts["clipDuration"]
        if fileOpts["fps"] != 0:
            sysCall += "-r %d " % fileOpts["fps"]

        sysCall += "%s" % fileOpts["outFile"]

    elif fileOpts["program"] == "convert":
        sysCall = "convert "

        if fileOpts["delay"] != 0:
            sysCall += "-delay 1x%d " % fileOpts["fps"]
        if fileOpts["optimizeLayers"] != "":
            sysCall += "-layers Optimize "
        if fileOpts["imgCount"] != 0:
            sysCall += "a{1..%d}.png " % fileOpts["imgCount"]
        else:
            sysCall += "%s " % fileOpts["inFile"]
        
        sysCall += "%s" % fileOpts["outFile"]

    ##  We add this to the end to send output/errors to /dev/null
    sysCall += " 2> /dev/null"

    return sysCall


def determine_if_valid(fileExtension):
    """
    This function runs through tuples of recognized file extensions to see if
    a given file has a valid extension.  If so, a true value is returned.  If
    not, an error is printed and a false value is returned.
    """
    if fileExtension in VID_EXT or fileExtension.swapcase() in VID_EXT:
        return True, "video"
    elif fileExtension in AUD_EXT or fileExtension.swapcase() in AUD_EXT:
        return True, "audio"
    elif fileExtension in IMG_EXT or fileExtension.swapcase() in IMG_EXT:
        return True, "image"
    else:
        return False, "none"

def no_extension(io, fileStr):
    """
    This prints an error telling the user what went wrong and exits the program.
    Specifically, if no extension was found on the input or output file, it
    exits.
    """
    print("Error:  No valid extension found on %s filename.  Skipping %s" %
            (io, fileStr))


def determine_extension_type(io, fileStr, options):
    """
    This function checks the extension of the input/output file and determines
    whether it is an image, audio or video file.
    """
    fileName = fileStr

    ##  Set initial value for 'valid'
    valid = False

    try:
        fileExt = fileName[fileName.rindex('.')+1:]
        valid, fileType = determine_if_valid(fileExt)
    except ValueError:
        if ((fileName == "goodgif") or (fileName == "bestgif") or
                (fileName == "gif") or (fileName == "smallgif")):
            fileExt = fileName
            fileType = fileName
            valid = True

        for e in VID_EXT:
            if fileName == e:
                fileExt = fileName
                fileType = "video"
                valid = True
        for e in AUD_EXT:
            if fileName == e:
                fileExt = fileName
                fileType = "audio"
                valid = True
        for e in IMG_EXT:
            if fileName == e:
                fileExt = fileName
                fileType = "image"
                valid = True

    if valid == False:
        no_extension(io, fileStr)

    return fileExt, valid, fileType


def image_to_image(inFile, inFileExt, outFile, outFileExt, options):
    """
    This function is the easiest.  It just runs the 'convert' ImageMagick
    program, which does all of the work.  Frankly, this is an entirely
    redundant feature of this program, but I'm including it just for
    completion's sake.
    """
    ##  First, we make the file extensions lowercase.
    iExt = inFileExt.lower()
    oExt = outFileExt.lower()

    ##  Check to see if both input file type and output file type are valid
    if iExt not in IMG_EXT:
        print("ERROR:  %s is not a valid image extension.  Skipping %s" %
                (inFileExt, inFile))
        return 0

    if oExt not in IMG_EXT:
        print("ERROR:  %s is not a valid image extension.  Skipping %s" %
                (outFileExt, inFile))
        return 0
    
    if options["verbose"] == True:
        print("Converting %s to %s" % (inFile, outFile))
    ##  We send stdout/stderr to /dev/null to prevent output from showing
    ##  on the screen as the script does its thing
    os.system("convert %s %s 2> /dev/null" % (inFile, outFile))

def find_frame_rate(inFile):
    """
    A function that uses ffprobe to find the frame rate of a video
    """
    ##  Execute ffprobe, piping out/err
    p = subprocess.Popen(["ffprobe","-show_streams",inFile], 
            stdout = subprocess.PIPE, stderr = subprocess.PIPE)

    out, err = p.communicate()
    
    outStr = str(out)

    ##  Here we seek out the r_frame_rate number, then slowly chip away
    ##  until we have the numbers we need
    fps_p1 = outStr.partition("r_frame_rate=")
    fps_p2 = fps_p1[2].partition("\\n")
    fps_p3 = fps_p2[0].partition('/')
    fps1 = fps_p3[0]
    fps2 = fps_p3[2]

    ##  Convert the strs to numbers, divide front part by rear part
    fps = int( int(fps1) / int(fps2) )


    return fps


def sanitize_filename(fileStr):
    """
    Fixes a filename to work properly on UNIX-likes
    """
    fileStr = fileStr.replace(" ","\ ",-1)

    return fileStr

def video_to_video(inFile, inFileExt, outFile, outFileExt, options):
    """
    This function uses ffmpeg to convert from one video format to another.
    """
    ##  Make file extensions lowercase
    iExt = inFileExt.lower()
    oExt = outFileExt.lower()

    ##  Check to see if input/output file formats are valid.
    if iExt not in VID_EXT:
        print("ERROR:  %s is not a valid video extension.  Skipping %s" %
                (inFileExt, inFile))
        return 0

    if oExt not in VID_EXT:
        print("ERROR:  %s is not a valid video extension.  Skipping %s" %
                (outFileExt, inFile))
        return 0

    ##  If verbosity is on, tell the user what's happening
    if options["verbose"] == True:
        print("Converting %s to %s..." % (inFile, outFile))

    ##  Find the frame rate of the video
    fps = find_frame_rate(inFile)

    if oExt == "mp4" or oExt == "m4v":
        os.system("ffmpeg -y -i %s -acodec copy -r %d %s 2> /dev/null" %
                (inFile, fps, outFile))
    elif oExt == "flv":
        os.system("ffmpeg -y -i %s -b:v 2000k -r %d %s 2> /dev/null" %
                (inFile, fps, outFile))
    elif oExt == "avi":
        os.system("ffmpeg -y -i %s -f avi -b:v 2000k -r %d %s 2> /dev/null" %
                (inFile, fps, outFile))
    elif oExt == "wmv":
        os.system("ffmpeg -y -i %s -vcodec wmv2 -b:v 2000k -acodec copy -r %d %s 2> /dev/null" %
                (inFile, fps, outFile))
    elif oExt == "mpg" or oExt == "mpeg" or oExt == "vob":
        os.system("ffmpeg -y -i %s -sameq -r %d %s 2> /dev/null" % (
            inFile, fps, outFile))
    elif oExt == "webm":
        os.system("ffmpeg -y -i %s -f webm -b:v 2000k -r %d %s 2> /dev/null" %
                (inFile, fps, outFile))
    elif oExt == "mov":
        os.system("ffmpeg -y -i %s -f mov -acodec copy -r %d %s 2> /dev/null" %
                (inFile, fps, outFile))
    elif oExt == "mkv":
        os.system("ffmpeg -y -i %s -r %d %s 2> /dev/null" % (inFile, fps, outFile))
    elif oExt == "ogv":
        os.system("ffmpeg -y -i %s -b:v 2000k -r %d %s 2> /dev/null" %
                (inFile, fps, outFile))


def make_animated_gif(inFile, outFile, options, quality="gif"):
    """
    This is the function used to make an animated gif.
    """
    ##  First, we check to make sure that the outFile name isn't just the
    ##  quality of the gif.  If it is, make it the same name as inFile
    if outFile in ("goodgif","bestgif","gif","smallgif"):
        outFile = inFile[:-4]

    ##  Next, we check to see that the outFile ends in .gif.  If not,
    ##  make it so.
    if outFile[-4:] != ".gif" and outFile[-4:] != ".GIF":
        outFile += ".gif"

    ##  Create working directory, move into it, create images, make gif,
    ##  remove images, copy gif back out, then remove directory
    workDir = "vert_working_dir"

    try:
        os.mkdir(workDir)
    except OSError:
        print("Error:  Cannot create directory.  Skipping %s" % (outFile))
        return 0

    os.chdir(workDir)

    ##  Link to video in parent directory
    os.link('../%s' % inFile, inFile)
    
    ##  And here begins the long, drawn-out process of finding the size of
    ##  the video.

    ##  Do all this crap to actually get output from a called program
    p = subprocess.Popen(["ffprobe", "-show_streams", inFile], 
            stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    out, err = p.communicate()

    ##  Turn the output (in this case, stdout) into a string
    outStr = str(out)

    ##  Partition string, find "width=", then grab what's after it
    p1 = outStr.partition("width=")
    xStr = p1[2].partition("\\n")[0]

    ##  Partition remainder from p1, then do the same to it
    p2 = p1[2].partition("height=")
    yStr = p2[2].partition("\\n")[0]

    ##  Turn both strings into integers
    resX = int(xStr)
    resY = int(yStr)

    ##  From here, we begin actually making the GIF

    ##  We size the gif according to quality.  Better == much higher file size

    if quality == "bestgif":
        gifX = resX
        gifY = resY
    elif quality == "goodgif":
        if resX/2 >= 200:
            gifX = resX/2
            gifY = resY/2
        else:
            gifX = resX
            gifY = resY
    elif quality == "gif":
        if resX/4 < 200:
            if resX/2 < 200:
                gifX = resX
                gifY = resY
            else:
                gifX = resX / 2
                gifY = resY / 2
        else:
            gifX = resX / 4
            gifY = resY / 4
    elif quality == "smallgif":
        gifX = resX
        gifY = resY
        while gifX > 200:
            gifX /= 2
            gifY /= 2
    
    ##  We find the frame rate of the video, and then set it according to
    ##  the quality selected.
    vidFPS = find_frame_rate(inFile)

    if quality == "bestgif":
        fps = vidFPS
    elif quality == "goodgif":
        if vidFPS < 20:
            fps = vidFPS
        else:
            fps = 20
    elif quality == "gif":
        if vidFPS > 15:
            fps = 15
        else:
            fps = vidFPS
    elif quality == "smallgif":
        if vidFPS > 10:
            fps = 10
        else:
            fps = vidFPS

    ##  And now, finally, the actual commands used
    outputStr = r"a%d.png"

    ##  Turn the clip into a series of PNG images

    ##  If verbosity is on, let the user know what's up.
    if options["verbose"] == True:
        print("Converting %s to an animated gif" % inFile)

    ##  If the user has specific clip start/duration, use that
    if options["clipStart"] != 0:
        if options["clipDuration"] != 0:
            os.system("ffmpeg -y -i %s -r %d -s %dx%d -ss %d -t %d %s 2> /dev/null"%
                    (inFile, fps, gifX, gifY, options["clipStart"],
                        options["clipDuration"], outputStr))
        else:
            os.system("ffmpeg -y -i %s -r %d -s %dx%d -ss %d %s 2> /dev/null" %
                    (inFile, fps, gifX, gifY, options["clipStart"], outputStr))
    elif options["clipDuration"] != 0:
        os.system("ffmpeg -y -i %s -r %d -s %dx%d -t %d %s 2> /dev/null" %
                (inFile, fps, gifX, gifY, options["clipDuration"], outputStr))
    else:
        os.system("ffmpeg -y -i %s -r %d -s %dx%d %s 2> /dev/null" %
                (inFile, fps, gifX, gifY, outputStr))

    imgCount = 0

    ##  Count how many images were produced
    for i in os.listdir():
        if ".png" in i:
            imgCount += 1

    ##  Convert PNG images into animated gif
    os.system("convert -delay 1x%d -layers Optimize a{1..%d}.png %s 2> /dev/null" %
            (fps, imgCount, outFile))

    ##  Move final product back into parent directory
    if os.path.exists('../%s' % outFile):
        os.remove('../%s' % outFile)
    shutil.move('%s' % outFile, '..')

    ##  Get rid of the evidence
    for f in os.listdir():
        os.remove(f)

    ##  cd back up, then remove directory
    os.chdir('..')
    os.rmdir(workDir)


def video_to_images(inFile, inFileExt, outFile, outFileExt, options):
    """
    This function will turn a video clip into a series of images.  If 'gif' is
    attempted, it will create an animated gif of a clip.
    """
    ##  Change file extensions to lowercase
    iExt = inFileExt.lower()
    oExt = outFileExt.lower()

    ##  Add '_img_' to the file extension
    extLoc = outFile.rfind('.')
    outExt = outFile[extLoc:]
    outputStr = outFile.replace(outExt, '')
    outputStr += "_img_"
    outputStr += r"%d"
    outputStr += outExt

    ##  Make an animated gif
    if oExt == "gif":
        make_animated_gif(inFile, outFile, options, quality="gif")
    elif oExt in IMG_EXT or oExt.swapcase() in IMG_EXT:
        if options["verbose"] == True:
            print("Converting %s to a series of images..." % inFile)
        os.system("ffmpeg -y -i %s %s 2> /dev/null" % (inFile, outputStr))


def video_to_audio(inFile, inFileExt, outFile, outFileExt, options):
    """
    This function rips audio from video to a variety of formats.
    """
    ##  Make extension lowercase
    oExt = outFileExt.lower()

    ##  If verbosity is on, let the user know what's going on
    if options["verbose"] == True:
        print("Converting %s to %s" % (inFile, outFile))

    ##  Save to MP3
    if oExt == "mp3" or oExt == "wav":
        os.system("ffmpeg -y -i %s -vn -ac 2 -ar 44100 -f %s %s 2> /dev/null" %
                (inFile, oExt, outFile))
    elif oExt == "flac":
        os.system("ffmpeg -y -i %s %s 2> /dev/null" % (inFile, outFile))
    elif oExt == "ogg" or oExt == "oga":
        os.system("ffmpeg -y -i %s -vn -acodec libvorbis %s 2> /dev/null" %
                (inFile, outFile))
    elif oExt == "aiff":
        os.system("ffmpeg -y -i %s -vn -f aiff %s 2> /dev/null" % 
                (inFile, outFile))
    elif oExt == "ac3":
        os.system("ffmpeg -y -i %s -acodec ac3 %s 2> /dev/null" %
                (inFile, outFile))
    elif oExt == "spx" or oExt == "speex":
        os.system("ffmpeg -y -i %s -f ogg -vn -sn -acodec libspeex -ar 32000 %s 2> /dev/null" %
                (inFile, outFile))
    else:
        print("Error:  %s not currently support (or not a valid format)." % 
                (outFile))


def cannot_convert(inFile, inFileType, outFileType):
    """
    Generic "can't convert this" message.
    """
    print("Error:  Can't convert from %s to %s.  Skipping %s." %
            (inFileType, outFileType, inFile))



def verify_outfile_name(inFile, outFile, options):
    """
    Just making sure that the outFile isn't stupidly named.  If it is, we
    rename it to be the same as the inFile, except with a different extension.
    """
    ##  If the outfile doesn't contain an extension (i.e., IS the extension)
    if outFile not in ("goodgif","gif","smallgif","bestgif"):
        if '.' not in outFile:
            newOutFile = inFile[:-4] + ".%s" % outFile
            #if options["verbose"] == True:
            #    print("Renaming %s to %s" % (outFile, newOutFile))
            outFile = newOutFile

    return outFile



def conversion(inFile, outFile, options):
    """
    This function checks the i/o files and then determines if conversion is
    possible between the formats.  If not, the files are skipped.  If so,
    outside programs will be called (imagemagick or ffmpeg) to do the actual
    conversion.
    """
    inFileExt, inFileValid, inFileType = determine_extension_type("input",
            inFile, options)
    outFileExt,outFileValid,outFileType=determine_extension_type("output",
            outFile, options)

    if ((inFileValid == False) or (outFileValid == False)):
        return 0
    
    ##  Quick function to make sure the outfile isn't named stupidly.
    outFile = verify_outfile_name(inFile, outFile, options)

    ##  Here we employ a function that fixes filesnames
    inFile = sanitize_filename(inFile)
    outFile = sanitize_filename(outFile)

    if inFileType == "image":
        if outFileType == "image":
            image_to_image(inFile,  inFileExt, outFile, outFileExt, options)
        else:
            cannot_convert(inFile, inFileType, outFileType)
    elif inFileType == "video":
        if outFileType == "video":
            video_to_video(inFile, inFileExt, outFile, outFileExt, options)
        elif ((outFileType == "gif") or (outFileType == "goodgif") or
                (outFileType == "bestgif") or (outFileType == "smallgif")):
            make_animated_gif(inFile, outFile, options, outFileType)
        elif outFileType == "image":
            video_to_images(inFile, inFileExt, outFile, outFileExt, options)
        elif outFileType == "audio":
            video_to_audio(inFile, inFileExt, outFile, outFileExt, options)
        else:
            cannot_convert(inFile, inFileType, outFileType)
    elif inFileType == "audio":
        if outFileType == "audio":
            video_to_audio(inFile, inFileExt, outFile, outFileExt, options)
        else:
            cannot_convert(inFile, inFileType, outFileType)

def print_help():
    """
    Prints the help.
    """
    helpStr = """
Vert is a quick and dirty method of converting media files to other types of
media.  Video files can be converted to other video formats, audio files, a
series of images or animated gifs.  Image files can be converted to other image
formats, and audio files can be converted to other audio formats.

Usage:
    vert [args] INPUT OUTPUT.format

Args:
    -h or --help:       Display help
    -v or --verbose:    Verbose output
    aaa
    """
    print(helpStr)

def print_version():
    """
    Prints the version.
    """
    versionStr = """
    Program:    vert.py, version 0.8dev
    Author:     James Hendrie (hendrie dot james at gmail dot com)
    """
    print(versionStr)

def set_options():
    """
    Set the initial options for the script
    """
    ##  Initial options dictionary
    options = {"verbose":False, "silent":False, "safe":False, "clipStart":0,
            "clipDuration":0, "inFileExtension":"", "inFileFormat":"",
            "outFileExtension":"","outFileFormat":"", "inFile":"","outFile":"",
            "videoBitrate":"","audioCodec":"","noVideo":"","fps":"",
            "inFileType":"","outFileType":"","videoCodec":"",
            "program":""}

    ##  Return the options dict
    return options


def get_options(options, argsList):
    """
    Check the input for various arguments and set options accordingly.
    """
    ##  We use this to know what arguments to remove later
    removeArgs = []

    for arg in argsList:
        if (arg == "--help") or (arg == "-h"):
            print_help()
            sys.exit()
        elif (arg == "--version"):
            print_version()
            sys.exit()
        elif (arg == "--verbose") or (arg == "-v"):
            options["verbose"] = True
            argsList.remove(arg)
        elif (arg == "--silent") or (arg == "-s"):
            options["silent"] = True
            argsList.remove(arg)
        elif (arg == "--safe"):
            options["safe"] = True
            argsList.remove(arg)
        elif (arg == "-ss"):
            try:
                options["clipStart"] = int(argsList[argsList.index(arg) + 1])

                removeArgs.append(argsList[argsList.index(arg) + 1])
                removeArgs.append(arg)
            except ValueError:
                print("Error:  %s is not a valid starting position.  Aborting."%
                        argsList[argsList.index(arg)+1])
                sys.exit()
        elif (arg == "-t"):
            try:
                options["clipDuration"]= int(argsList[argsList.index(arg) + 1])

                #argsList.remove(argsList[argsList.index(arg) + 1])
                #argsList.remove(arg)
                removeArgs.append(argsList[argsList.index(arg) + 1])
                removeArgs.append(arg)
            except ValueError:
                print("Error:  %s is not a valid clip duration  Aborting." %
                        argsList[argsList.index(arg)+1])
                sys.exit()

    for arg in removeArgs:
        argsList.remove(arg)

    return options, argsList


def get_files(argsList, batchMode = False):
    """
    This function creates a dictionary of input/output files according to the
    arguments passed to it and the mode selected.  If batchMode is True, then
    the list of files will all be set to conver to the last argument (format).
    Otherwise, it's assumed the user is converting from one file to another.
    """
    ##  Create files dictionary
    filesDict = {}

    if batchMode == True:
        ##  For all arguments except for the last one (the format)
        for a in argsList[:-1]:
            ##  Set all the files' chosen format as the last argument
            filesDict[a] = argsList[-1]
    else:
        ##  Input/output is easy to set in this circumstance
        filesDict[argsList[0]] = argsList[1]

    return filesDict


def main():
    """
    The main function.  Grabs stuff from stdin (input/output/args) and deals
    with it.
    """

    ##  Generate an options dictionary
    options = set_options()

    ##  Get additional options from the command line, as well as some args
    ##  (initially, we pass all args except for the program name)
    options, argsList = get_options(options, sys.argv[1:])

    ##  At this point, we determine if the user is doing 'batch' processing or
    ##  just converting one file to another format.
    if len(argsList) == 2:
        filesDict = get_files(argsList)
    elif len(argsList) > 2:
        filesDict = get_files(argsList, True)
    elif len(argsList) < 2:
        print("Error:  Too few arguments/files.  Aborting.")
        sys.exit()

    ##  Go through the dictionary, converting files
    for key in filesDict:
        conversion(key, filesDict[key], options)


if __name__ == "__main__":
    main()
